// ------------------------------------------------------------------------ //
//                                                                          //
// File:    PreLang.cpp                                                     //
// Author:  Dirk Schwarzmann                                                //
// Contact: Dirk.Schwarzmann@gmx.de                                         //
//                                                                          //
// Version: 1.0                                                             //
// Date:    15th Febuary 2001                                               //
//                                                                          //
// Purpose:                                                                 //
//          Reads an exported reg file that contains the tree of            //
//          [HKLM\SYSTEM\CurrentControlSet\Control\Keyboard Layouts]        //
//          and shows all available keyboard layouts on screen.             //
//          The user can select one layout which is in return exported to   //
//          to a file for the purpose of setting the keyboard language      //
//          layout at login state.                                          //
//                                                                          //
//          The program takes a file name to read as the first parameter    //
//          and a file name for writing the out as a second one.            //
//          The third one indicates wether to write Windows 2000 output or  //
//          output for Windows 9x                                           //
//                                                                          //
//          call: prelang <infile> <outfile> -win2k | -win9x                //
//                                                                          //
//          output: a file named <outfile> containing the selection or an   //
//                  ERRORLEVEL in case some error occured:                  //
//                  1: input file not found                                 //
//                  2: layout file not found                                //
//                                                                          //
// ------------------------------------------------------------------------ //


#include <iostream>
#include <fstream>
#include <string>
using namespace std;

// Datenstruktur fuer die Language-Daten
struct Layout {
  string id;    // ID-Nummer, z.B. 00000407 fuer Deutsch
  string file;  // Name der Datei, in der das Layout gespeichert ist
  string text;  // Beschreibender Text, wie er in der Systemsteuerung auftaucht
};

ifstream inFile; // Das Objekt fuer die einzulesende Datei

const int lineLen = 400;         // Max. Laenge einer Zeile
char* line = new char[lineLen];  // Eine gelesene Zeile


bool initFile(char* fname) {
  // Bereitet die in 'fname' angegebene Datei auf das Lesen vor. Die Routine
  // prueft, ob die Datei UNICODE-codiert ist oder nicht und gibt den
  // entsprechenden Wahrheitswert zurueck: true, wenn UNICODE, false sonst.
  // Auerdem bleibt die Datei lese-geoeffnet, wobei der Filepointer je nach
  // Codierung auf dem ersten zu lesenden Zeichen steht.
  // Fuer weiteres Einlesen muss daher die Codierung bekannt sein!

  char c;
  
  inFile.open(fname, ios::in);
  inFile.get(c);

  if (c == '\xFF') {

    // Es ist ein UNICODE-Datei
    inFile.get(c); // Ueberlies auch das zweite Zeichen

    return true;

  } else {

    // Es ist eine ASCII-Datei
    // Wir haben schon zuviel eingelesen. Zum Ruecksetzen des Filepointer wird
    // die Datei geschlossen und wieder geoeffnet:
    inFile.close();
    inFile.open(fname, ios::in);

    return false;

  }

} // bool initFile(char*)

string readFileLine(bool isUNICODE) {
  // Liest eine Zeile einer Textdatei.
  // In 'isUNICODE' wird bestimmt, ob eine normale ASCII-Zeile oder UNICODE
  // erwartet wird.

  string result;

  if (isUNICODE) {

    // Es handelt sich um eine UNICODE-Zeile.
    // Im Grunde wird einfach jedes zweite Zeichen (Null-Zeichen) heraus-
    // gefiltert.
    char c;

    while (!inFile.eof()) {
      inFile.get(c); // Erstes echtes Zeichen

      if (c == '\x0D') {

        // Zeilenumbruch beginnt. Besteht aus 4 Zeichen: 0D 00 0A 00
        inFile.get(c); // Die folgenden drei Zeichen koennen ignoriert
        inFile.get(c); // werden; sie sind bekannt. Danach ist die Zeile
        inFile.get(c); // dann zu Ende.

        return result;
      }

      result += c;   // Fuege das verwertbare Zeichen dem Ergebnis hinzu
      inFile.get(c); // und ignoriere den naechsten Null-Character

    } // while (!inFile.eof())

    return result;

  } else { // if (isUNICODE)

    // Es handelt sich um eine ASCII-Zeile:
    inFile.getline(line, lineLen);
    result = line;
    return result;

  } // else if (isASCII)

} // string readFileLine(bool)

string str2upper(string str) {
  // Konvertiert einen String in Grossbuchstaben

  for (int j=0; j<str.length(); j++) {

    if (str[j] >= 'a' && str[j] <= 'z') {
      str[j] = toupper(str[j]);
    }
  } // for (int j=0; ...)

  return str;

} // string str2upper(string)


void main(int argc, char* argv[]) {

  char inName[200];  // Der Name der Eingabedatei

  string result;     // Dieselbe Zeile, konvertiert als String

  bool unicode;      // Flag, ob die Datei UNICODE-codiert ist

  bool writeWin2k;   // Schreibe Windows 2000 oder Windows 9x Struktur

  const string rnode = "[HKEY_LOCAL_MACHINE\\SYSTEM\\CURRENTCONTROLSET\\CONTROL\\KEYBOARD LAYOUTS\\";
  const string rfile = "LAYOUT FILE";
  const string rtext = "LAYOUT TEXT";

  Layout lang[100]; // Das Array haelt alle verfuegbaren Layouts

  int idx = 0;

  if (argc < 4) {

    // Zuwenig Parameter angegeben, aufhoeren:
    exit(1);
  }

  // Alles in Ordnung
  strcpy(inName, argv[1]);

  inFile.open(inName, ios::in);

  if (!inFile) {

    // Konnte Datei nicht oeffnen
    // Laesst sich per Batch ueber ERRORLEVEL abfragen
    inFile.close();
    exit(1);

  } else {

    inFile.close();

    unicode = initFile(argv[1]);

    // Lies die erste Zeile
    result = str2upper(readFileLine(unicode));

  } // else if (!inFile)

  // Entscheide, welche Registry-Struktur geschrieben werden soll. Per default
  // wird Windows 2000 geschrieben:
  if (str2upper(argv[3]) == "-WIN9X") {

    writeWin2k = false;

  } else {

    writeWin2k = true;

  }

  ////////////////////////////////////////////////////////////////////////////
  // Lies alle Daten ein und lege sie im Array ab
  ////////////////////////////////////////////////////////////////////////////
  while (!inFile.eof()) {
    
    // Pruefe, ob die Zeile einen Language-Knoten enthaelt:
    if (result.find(rnode) != string::npos) {
    
      // Ja; entferne alles bis auf seine ID-Nummer:
      // vorne...
      result.replace(0, rnode.length(), "");
      // ...und hinten
      result.replace(result.length()-1, result.length(), "");
        
      // Speichere die ID:
      lang[idx].id = result;
    
      // Solange die naechste Zeile kein weiterer Language-Knoten ist...
      while (!inFile.eof() && result.find(rnode) == string::npos) {
    
        // Lies die naechsten Zeilen, die Text und File enthalten:
        result = str2upper(readFileLine(unicode));

        // Ueberspringe leere Zeilen:
        if (result.length() > 0) {
    
          // "File"-Eintrag gefunden?
          if (result.find(rfile) != string::npos) {
    
            // Speichere den Dateinamen:
            lang[idx].file = result;
            lang[idx].file.replace(0,15, "");
            lang[idx].file.replace(lang[idx].file.length()-1, lang[idx].file.length(), "");
          }
    
          // "Text"-Eintrag gefunden?
          if (result.find(rtext) != string::npos) {
    
            // Speichere den Dateinamen:
            lang[idx].text = result;
            lang[idx].text.replace(0,15, "");
            lang[idx].text.replace(lang[idx].text.length()-1, lang[idx].text.length(), "");
          }
    
        } // if (result.length() > 0)
    
      } // while (!inFile.eof() && result.find(rnode != string::npos)
    
      idx++; // Naechster Eintrag
    
    } else { // if (result.find(rnode) != string::npos)
    
      result = str2upper(readFileLine(unicode));
    
    }

  } // while (!inFile.eof());

  idx--; // Index ist um 1 zu hoch

  inFile.close();

  ////////////////////////////////////////////////////////////////////////////
  // Beginne Anzeige
  ////////////////////////////////////////////////////////////////////////////
  int language = 0;  // Noch keine Sprache gewaehlt
  int start = 1;     // Ab welchem Index die Layouts angezeigt werden sollen
  string userInput;

  while (language == 0) {

    cout << "Welches Sprachlayout moechten Sie beim Einloggen standardmaessig haben?\n\n";

    for (int j=start; j<start+20; j++) {
      cout << j << ".\t" << lang[j-1].text << endl;
    }
    
    cout << "\nGeben Sie die jeweilige Zahl ein oder 'w' fuer weitere, 'v' fuer vorige\n";

    cin >> userInput;

    // Weitere Layouts anzeigen:
    if (userInput == "w" && start+20 <= idx) {
        start += 20;
    }

    // Vorige Layouts anzeigen:
    if (userInput == "v" && start-20 >= 0) {
        start -= 20;
    }

    // Wandle die Eingabe in eine Zahl (Auswahl ueberpruefen):
    language = atoi(userInput.c_str());

    // Wenn die Auswahl ausserhalb des gueltigen Bereiches liegt...
    if (language < 1 || language > idx+1) {

      // ...verwirf sie:
      language = 0;
    }

  } // while (language == 0)

  // Die Nummer der ausgewaehlten Sprache ist um 1 zu hoch, denn das Array
  // beginnt bei 0:
  language--;

  ////////////////////////////////////////////////////////////////////////////
  // Generiere Output
  ////////////////////////////////////////////////////////////////////////////

  // Teste, ob das ausgewaehlte Layout installiert ist:
  char* lfile = getenv("WINDIR");
  
  if (writeWin2k) {
    strcat(lfile, "\\system32\\");
  } else {
  	strcat(lfile, "\\system\\");
  }
  strcat(lfile, lang[language].file.c_str());

  ifstream sysFile;
  sysFile.open(lfile, ios::in);

  if (!sysFile) {

    // Konnte Treiber nicht oeffnen
    // Laesst sich per Batch ueber ERRORLEVEL abfragen
    sysFile.close();
    exit(2);
  }

  sysFile.close();

  // Es ist alles in Ordnung, wir koennen die Ausgabedatei anlegen:

  ofstream outFile; // Das Objekt fuer die einzulesende Datei

  outFile.open(argv[2], ios::out);

  outFile.write("REGEDIT4\n\n", 10);
  outFile.write("[HKEY_USERS\\.DEFAULT\\Keyboard Layout\\Preload", 44);

  // Unterschiedliche Ausgabe fuer Windows 2000 bzw. Windows 9x:
  if (writeWin2k) {

    outFile.write("]\n", 2);
    outFile.write("\"1\"=\"", 5);

  } else {

    outFile.write("\\1]\n", 4);
    // Ein leerer String steht fuer den "(Standard)" Eintrag:
    outFile.write("\"\"=\"", 4);

  }

  outFile.write(lang[language].id.c_str(), lang[language].id.length());
  outFile.write("\"\n", 2);

  outFile.close();

  exit(0);
}; // void main
